@@ -24,7 +24,6 @@
 #endif
 
 #include <glib/gi18n.h>
-#include <glade/glade.h>
 #include <gnome-keyring.h>
 
 #include "vinagre-tab.h"
@@ -68,7 +67,8 @@ enum
 {
   PROP_0,
   PROP_CONN,
-  PROP_WINDOW
+  PROP_WINDOW,
+  PROP_TOOLTIP
 };
 
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -137,6 +137,8 @@ vinagre_tab_get_property (GObject    *object,
       case PROP_WINDOW:
         g_value_set_object (value, tab->priv->window);
 	break;
+      case PROP_TOOLTIP:
+	g_value_take_string (value, vinagre_tab_get_tooltip (tab));
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	break;			
@@ -218,6 +220,156 @@ default_get_extra_title (VinagreTab *tab)
   return NULL;
 }
 
+static void
+menu_position (GtkMenu    *menu,
+	       gint       *x,
+	       gint       *y,
+	       gboolean   *push_in,
+	       VinagreTab *tab)
+{
+  int width, height;
+  GdkWindow *window;
+
+  window = gtk_widget_get_window (tab->priv->toolbar);
+
+  gdk_window_get_origin (window, x, y);
+  gdk_drawable_get_size (window, &width, &height);
+
+  *push_in = TRUE;
+  *y += height;
+}
+
+static void
+open_connection_cb (GtkMenuItem *item,
+		    VinagreTab  *tab)
+{
+  vinagre_window_set_active_tab (tab->priv->window, tab);
+}
+
+static void
+active_connections_button_clicked  (GtkToolButton *button,
+				    VinagreTab    *tab)
+{
+  GSList            *connections, *l;
+  VinagrePlugin     *plugin;
+  VinagreConnection *conn;
+  GtkWidget         *menu, *item, *image;
+  gchar             *str, *label;
+
+  menu = gtk_menu_new ();
+
+  connections = vinagre_notebook_get_tabs (tab->priv->nb);
+  for (l = connections; l; l = l->next)
+    {
+      conn = VINAGRE_TAB (l->data)->priv->conn;
+      plugin = vinagre_plugins_engine_get_plugin_by_protocol (vinagre_plugins_engine_get_default (),
+							      vinagre_connection_get_protocol (conn));
+      item = gtk_image_menu_item_new_with_label ("");
+      image = gtk_image_new_from_icon_name (vinagre_plugin_get_icon_name (plugin),
+					    GTK_ICON_SIZE_MENU);
+      gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
+
+      label = vinagre_connection_get_best_name (conn);
+      if ((l->data == tab) && (GTK_IS_LABEL (GTK_BIN (item)->child)))
+	{
+	  str = g_strdup_printf ("<b>%s</b>", label);
+	  gtk_label_set_use_markup (GTK_LABEL (GTK_BIN (item)->child), TRUE);
+	  g_free (label);
+	  label = str;
+	}
+      gtk_menu_item_set_label (GTK_MENU_ITEM (item), label);
+      g_free (label);
+
+      gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+      g_signal_connect (item, "activate", G_CALLBACK (open_connection_cb), l->data);
+    }
+
+  gtk_widget_show_all (menu);
+  gtk_menu_popup (GTK_MENU (menu),
+		  NULL,
+		  NULL,
+		  (GtkMenuPositionFunc) menu_position,
+		  tab,
+		  0,
+		  gtk_get_current_event_time ());
+}
+
+static void
+close_button_clicked (GtkToolButton *button,
+		      VinagreTab    *tab)
+{
+  vinagre_notebook_close_tab (tab->priv->nb, tab);
+}
+
+static void
+fullscreen_button_clicked (GtkToolButton *button,
+			   VinagreTab    *tab)
+{
+  vinagre_window_toggle_fullscreen (tab->priv->window);
+}
+
+static void
+setup_layout (VinagreTab *tab)
+{
+  GtkWidget  *button;
+  gchar      *str;
+
+  tab->priv->toolbar = gtk_toolbar_new ();
+  gtk_toolbar_set_show_arrow (GTK_TOOLBAR (tab->priv->toolbar), FALSE);
+  GTK_WIDGET_SET_FLAGS (tab->priv->toolbar, GTK_NO_SHOW_ALL);
+
+  gtk_toolbar_set_style (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOLBAR_BOTH_HORIZ);
+
+  /* Close connection */
+  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Close connection"));
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (close_button_clicked), tab);
+
+  /* Connection name/menu */
+  str = vinagre_connection_get_best_name (tab->priv->conn);
+  button = GTK_WIDGET (gtk_tool_button_new (NULL, str));
+  g_free (str);
+
+  str = vinagre_connection_get_string_rep (tab->priv->conn, TRUE);
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), str);
+  g_free (str);
+
+  gtk_tool_item_set_is_important (GTK_TOOL_ITEM (button), TRUE);
+  g_signal_connect (button, "clicked", G_CALLBACK (active_connections_button_clicked), tab);
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+
+  /* Leave fullscreen */
+  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_LEAVE_FULLSCREEN));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Leave fullscreen"));
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (fullscreen_button_clicked), tab);
+
+  tab->priv->layout = ViewAutoDrawer_New ();
+  ViewAutoDrawer_SetActive (VIEW_AUTODRAWER (tab->priv->layout), FALSE);
+  ViewOvBox_SetOver (VIEW_OV_BOX (tab->priv->layout), tab->priv->toolbar);
+  ViewOvBox_SetUnder (VIEW_OV_BOX (tab->priv->layout), tab->priv->scroll);
+  ViewAutoDrawer_SetOffset (VIEW_AUTODRAWER (tab->priv->layout), -1);
+  ViewAutoDrawer_SetFill (VIEW_AUTODRAWER (tab->priv->layout), FALSE);
+
+  gtk_box_pack_end (GTK_BOX(tab), tab->priv->layout, TRUE, TRUE, 0);
+  gtk_widget_show_all (GTK_WIDGET (tab));
+}
+
+static void
+vinagre_tab_constructed (GObject *object)
+{
+  VinagreTab *tab = VINAGRE_TAB (object);
+
+  if (G_OBJECT_CLASS (vinagre_tab_parent_class)->constructed)
+    G_OBJECT_CLASS (vinagre_tab_parent_class)->constructed (object);
+
+  setup_layout (tab);
+}
+
 static void 
 vinagre_tab_class_init (VinagreTabClass *klass)
 {
@@ -226,6 +378,7 @@ vinagre_tab_class_init (VinagreTabClass *klass)
   object_class->dispose  = vinagre_tab_dispose;
   object_class->get_property = vinagre_tab_get_property;
   object_class->set_property = vinagre_tab_set_property;
+  object_class->constructed = vinagre_tab_constructed;
 
   klass->impl_get_tooltip = NULL;
   klass->impl_get_screenshot = NULL;
@@ -258,6 +411,17 @@ vinagre_tab_class_init (VinagreTabClass *klass)
 							G_PARAM_STATIC_NICK |
 							G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (object_class,
+				   PROP_TOOLTIP,
+				   g_param_spec_string ("tooltip",
+							"Tooltip",
+							"The tooltip of this tab",
+							NULL,
+							G_PARAM_READABLE |
+							G_PARAM_STATIC_NAME |
+							G_PARAM_STATIC_NICK |
+							G_PARAM_STATIC_BLURB));
+
   signals[TAB_CONNECTED] =
 		g_signal_new ("tab-connected",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -338,88 +502,6 @@ vinagre_tab_add_recent_used (VinagreTab *tab)
   g_slice_free (GtkRecentData, data);
 }
 
-static void
-close_button_clicked (GtkToolButton *button,
-		      VinagreTab    *tab)
-{
-  vinagre_notebook_close_tab (tab->priv->nb, tab);
-}
-
-static void
-minimize_button_clicked (GtkToolButton *button,
-			 VinagreTab    *tab)
-{
-  gtk_window_iconify (GTK_WINDOW (tab->priv->window));
-}
-
-static void
-fullscreen_button_clicked (GtkToolButton *button,
-			   VinagreTab    *tab)
-{
-  vinagre_window_toggle_fullscreen (tab->priv->window);
-}
-
-static void
-screenshot_button_clicked (GtkToolButton *button,
-			   VinagreTab    *tab)
-{
-  vinagre_tab_take_screenshot (tab);
-}
-
-static void
-setup_layout (VinagreTab *tab)
-{
-  GtkWidget  *button;
-
-  tab->priv->toolbar = gtk_toolbar_new ();
-  GTK_WIDGET_SET_FLAGS (tab->priv->toolbar, GTK_NO_SHOW_ALL);
-
-  gtk_toolbar_set_style (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOLBAR_BOTH_HORIZ);
-
-  /* Close connection */
-  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE));
-  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Close connection"));
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-  g_signal_connect (button, "clicked", G_CALLBACK (close_button_clicked), tab);
-
-  /* Leave fullscreen */
-  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_LEAVE_FULLSCREEN));
-  g_object_set (button, "is-important", TRUE, NULL);
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-  g_signal_connect (button, "clicked", G_CALLBACK (fullscreen_button_clicked), tab);
-
-  /* Minimize window */
-  button = GTK_WIDGET (gtk_tool_button_new (NULL, NULL));
-  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Minimize window"));
-  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "view-minimize");
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-  g_signal_connect (button, "clicked", G_CALLBACK (minimize_button_clicked), tab);
-
-  /* Space */
-  button = GTK_WIDGET (gtk_separator_tool_item_new ());
-  gtk_separator_tool_item_set_draw (GTK_SEPARATOR_TOOL_ITEM (button), FALSE);
-  gtk_tool_item_set_expand (GTK_TOOL_ITEM (button), TRUE);
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-
-  /* Screenshot */
-  button = GTK_WIDGET (gtk_tool_button_new (NULL, NULL));
-  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "applets-screenshooter");
-  gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), _("Take screenshot"));
-  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Take screenshot"));
-  g_signal_connect (button, "clicked", G_CALLBACK (screenshot_button_clicked), tab);
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-
-  tab->priv->layout = ViewAutoDrawer_New ();
-  ViewAutoDrawer_SetActive (VIEW_AUTODRAWER (tab->priv->layout), FALSE);
-  ViewOvBox_SetOver (VIEW_OV_BOX (tab->priv->layout), tab->priv->toolbar);
-  ViewOvBox_SetUnder (VIEW_OV_BOX (tab->priv->layout), tab->priv->scroll);
-}
-
 static void
 vinagre_tab_init (VinagreTab *tab)
 {
@@ -435,11 +517,6 @@ vinagre_tab_init (VinagreTab *tab)
 				  GTK_POLICY_AUTOMATIC);
   gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (tab->priv->scroll),
 				       GTK_SHADOW_NONE);
-
-  setup_layout (tab);
-
-  gtk_box_pack_end (GTK_BOX(tab), tab->priv->layout, TRUE, TRUE, 0);
-  gtk_widget_show_all (GTK_WIDGET (tab));
 }
 
 GtkWidget *
@@ -673,14 +750,60 @@ vinagre_tab_remove_from_notebook (VinagreTab *tab)
   vinagre_notebook_close_tab (tab->priv->nb, tab);
 }
 
+static void
+add_if_writable (GdkPixbufFormat *data, GSList **list)
+{
+  if (gdk_pixbuf_format_is_writable (data))
+    *list = g_slist_prepend (*list, data);
+}
+
+static GSList *
+get_supported_image_formats (void)
+{
+  GSList *formats = gdk_pixbuf_get_formats ();
+  GSList *writable_formats = NULL;
+
+  g_slist_foreach (formats, (GFunc)add_if_writable, &writable_formats);
+  g_slist_free (formats);
+
+  return writable_formats;
+}
+
+static void
+filter_changed_cb (GObject *object, GParamSpec *pspec, VinagreTab *tab)
+{
+  GtkFileFilter  *filter;
+  gchar          *extension, *filename, *basename, *newbase;
+  GtkFileChooser *chooser = GTK_FILE_CHOOSER (object);
+  int            i;
+
+  filter = gtk_file_chooser_get_filter (chooser);
+  extension = g_object_get_data (G_OBJECT (filter), "extension");
+
+  filename = gtk_file_chooser_get_filename (chooser);
+  basename = g_path_get_basename (filename);
+  for (i = strlen (basename)-1; i>=0; i--)
+    if (basename[i] == '.')
+      break;
+  basename[i] = '\0';
+  newbase = g_strdup_printf ("%s.%s", basename, extension);
+
+  gtk_file_chooser_set_current_name (chooser, newbase);
+
+  g_free (filename);
+  g_free (basename);
+  g_free (newbase);
+}
+
 void
 vinagre_tab_take_screenshot (VinagreTab *tab)
 {
   GdkPixbuf     *pix;
   GtkWidget     *dialog;
   GString       *suggested_filename;
-  gchar         *filename, *name;
-  GtkFileFilter *filter;
+  gchar         *name;
+  GtkFileFilter *filter, *initial_filter;
+  GSList        *formats, *l;
 
   g_return_if_fail (VINAGRE_IS_TAB (tab));
 
@@ -693,12 +816,6 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
       return;
     }
 
-  filename = NULL;
-  name = vinagre_connection_get_best_name (tab->priv->conn);
-  suggested_filename = g_string_new (NULL);
-  g_string_printf (suggested_filename, _("Screenshot of %s"), name);
-  g_string_append (suggested_filename, ".png");
-
   dialog = gtk_file_chooser_dialog_new (_("Save Screenshot"),
 				      GTK_WINDOW (tab->priv->window),
 				      GTK_FILE_CHOOSER_ACTION_SAVE,
@@ -706,25 +823,71 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
 				      GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
 				      NULL);
   gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
-  gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), suggested_filename->str);
   gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
 
-  filter = gtk_file_filter_new ();
-  gtk_file_filter_set_name (filter, _("Supported formats"));
-  gtk_file_filter_add_pixbuf_formats (filter);
-  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
+  name = vinagre_connection_get_best_name (tab->priv->conn);
+  suggested_filename = g_string_new (NULL);
+  g_string_printf (suggested_filename, _("Screenshot of %s"), name);
+  g_string_append (suggested_filename, ".png");
+  g_free (name);
+  gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), suggested_filename->str);
+  g_string_free (suggested_filename, TRUE);
+
+  formats = get_supported_image_formats ();
+  for (l = formats; l; l = l->next)
+    {
+      GdkPixbufFormat *data = (GdkPixbufFormat *)l->data;
+      gchar **exts;
+      int i;
+
+      filter = gtk_file_filter_new ();
+
+      name = gdk_pixbuf_format_get_description (data);
+      gtk_file_filter_set_name (filter, name);
+      g_free (name);
+
+      exts = gdk_pixbuf_format_get_extensions (data);
+      g_object_set_data_full (G_OBJECT (filter), "extension", g_strdup (exts[0]), g_free);
+      if (strcmp (exts[0], "png") == 0)
+	initial_filter = filter;
+
+      for (i = 0; exts[i]; i++)
+	{
+	  name = g_strdup_printf ("*.%s", exts[i]);
+	  gtk_file_filter_add_pattern (filter, name);
+	  g_free (name);
+	}
+      g_strfreev (exts);
+
+      gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
+    }
+  g_slist_free (formats);
+
+  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), initial_filter);
+  g_signal_connect (dialog, "notify::filter", G_CALLBACK (filter_changed_cb), tab);
 
   if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
     {
-      filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
-      gdk_pixbuf_save (pix, filename, "png", NULL, NULL);
+      GError *error = NULL;
+      gchar *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
+
+      filter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
+      name = g_object_get_data (G_OBJECT (filter), "extension");
+      if (!name)
+	name = "png";
+
+      if (!gdk_pixbuf_save (pix, filename, name, &error, NULL))
+	{
+	  vinagre_utils_show_error (_("Error saving screenshot"),
+				    error->message,
+				    GTK_WINDOW (tab->priv->window));
+	  g_error_free (error);
+	}
       g_free (filename);
   }
 
   gtk_widget_destroy (dialog);
   g_object_unref (pix);
-  g_string_free (suggested_filename, TRUE);
-  g_free (name);
 }
 
 const GSList *
@@ -732,7 +895,7 @@ vinagre_tab_get_always_sensitive_actions (VinagreTab *tab)
 {
   g_return_val_if_fail (VINAGRE_IS_TAB (tab), NULL);
 
-  VINAGRE_TAB_GET_CLASS (tab)->impl_get_always_sensitive_actions (tab);
+  return VINAGRE_TAB_GET_CLASS (tab)->impl_get_always_sensitive_actions (tab);
 }
 
 const GSList *
@@ -740,7 +903,7 @@ vinagre_tab_get_connected_actions (VinagreTab *tab)
 {
   g_return_val_if_fail (VINAGRE_IS_TAB (tab), NULL);
 
-  VINAGRE_TAB_GET_CLASS (tab)->impl_get_connected_actions (tab);
+  return VINAGRE_TAB_GET_CLASS (tab)->impl_get_connected_actions (tab);
 }
 
 const GSList *
@@ -748,7 +911,7 @@ vinagre_tab_get_initialized_actions (VinagreTab *tab)
 {
   g_return_val_if_fail (VINAGRE_IS_TAB (tab), NULL);
 
-  VINAGRE_TAB_GET_CLASS (tab)->impl_get_initialized_actions (tab);
+  return VINAGRE_TAB_GET_CLASS (tab)->impl_get_initialized_actions (tab);
 }
 
 static void
@@ -768,4 +931,20 @@ vinagre_tab_free_actions (GSList *actions)
   g_slist_free (actions);
 }
 
+const gchar *
+vinagre_tab_get_icon_name (VinagreTab *tab)
+{
+  const gchar   *protocol;
+  VinagrePlugin *plugin;
+
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), NULL);
+
+  protocol = vinagre_connection_get_protocol (tab->priv->conn);
+  plugin = g_hash_table_lookup (vinagre_plugin_engine_get_plugins_by_protocol (vinagre_plugins_engine_get_default ()),
+				protocol);
+  g_return_val_if_fail (plugin != NULL, NULL);
+
+  return vinagre_plugin_get_icon_name (plugin);
+}
+
 /* vim: set ts=8: */
